//-------------------------------------------------------------------------------------------------------
//
// MCP2515 - CAN Controller
//
//-------------------------------------------------------------------------------------------------------


#include "Config.h"

#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <inttypes.h>

#include "MCP2515.h"
#include "Error.h"
#include "SPI.h"
#include "EEPROM.h"


//-------------------------------------------------------------------------------------------------------
// globale Variablen

volatile uint8_t MCP2515_tel_empfangen;		// pro empfangenem Telegramm erhhung um 1


//-------------------------------------------------------------------------------------------------------
// CAN Interrupt: Telegramm Empfangen
ISR(INT2_vect)
{
	MCP2515_tel_empfangen ++;
}


//-------------------------------------------------------------------------------------------------------
// MSP2515 Register schreiben

void F_MCP2515_reg_schreiben(uint8_t Adresse, uint8_t Daten)
{
	F_SPI_Reg_MCP2515();
	
	RESET(MCP2515_SS);
	
	F_SPI_senden(SPI_WRITE);
	F_SPI_senden(Adresse);
	F_SPI_senden(Daten);
	
	SET(MCP2515_SS);
}


//-------------------------------------------------------------------------------------------------------
// MSP2515 Register lesen

uint8_t F_MCP2515_reg_lesen(uint8_t Adresse)
{
	uint8_t Daten;
  
	F_SPI_Reg_MCP2515();
	
	RESET(MCP2515_SS);
	
	F_SPI_senden(SPI_READ);
	F_SPI_senden(Adresse);
	F_SPI_senden(0xFF);	
	Daten = F_SPI_lesen();	
	
	SET(MCP2515_SS);
	
	return Daten;
}


//-------------------------------------------------------------------------------------------------------
// einzelne Bits im MCP2515 ndern

void F_MCP2515_modify(uint8_t Adresse, uint8_t Maske, uint8_t Daten)
{
	F_SPI_Reg_MCP2515();
	
	RESET(MCP2515_SS);
	
	F_SPI_senden(SPI_BIT_MODIFY);
	F_SPI_senden(Adresse);
	F_SPI_senden(Maske);
	F_SPI_senden(Daten);
	
	SET(MCP2515_SS);
}


//-------------------------------------------------------------------------------------------------------
// MCP2515 Status auslesen

uint8_t F_MCP2515_status(uint8_t Type)
{
	uint8_t Daten;
	
	F_SPI_Reg_MCP2515();
	
	RESET(MCP2515_SS);
	
	F_SPI_senden(Type);
	F_SPI_senden(0xFF);
	Daten = F_SPI_lesen();
	
	SET(MCP2515_SS);
	
	return Daten;
}


//-------------------------------------------------------------------------------------------------------
// SPI und MCP2515 initialisierung

bool F_MCP2515_init(void)
{
	// externer Interrupt (INT2)
	SET_INPUT(MCP2515_INT);
	SET(MCP2515_INT);
	EICRA |= (1<<ISC01);	// fallende Flanken
	EIMSK |= (1<<INT2);
	EIMSK &= ~((1<<INT0) | (1<<INT1));
	
	F_SPI_Reg_MCP2515();

	// MCP2515 reseten, danach ist er im Konfigurationsmodus
	RESET(MCP2515_SS);
	F_SPI_senden(SPI_RESET);
	SET(MCP2515_SS);

	// kurze Wartezeit zum Neustarten des MCP2515
	_delay_us(10);

	// CNF1..3 Register laden
	RESET(MCP2515_SS);
	F_SPI_senden(SPI_WRITE);
	F_SPI_senden(CNF3);
	
	F_SPI_senden(1<<PHSEG21);		// Bitrate 125 kbps at 16 MHz
	F_SPI_senden((1<<BTLMODE) | (1<<PHSEG11));
	F_SPI_senden((1<<BRP2) | (1<<BRP1) | (1<<BRP0));
	
	// Interrups aktivieren
	F_SPI_senden((1<<RX1IE) | (1<<RX0IE));
	SET(MCP2515_SS);
	
	// test if we could read back the value => is the chip accessible?
	if(F_MCP2515_reg_lesen(CNF1) != ((1<<BRP2) | (1<<BRP1) | (1<<BRP0))) 
	{
		return 0;
	}
	
	// deaktivate the RXnBF Pins (High Impedance State)
	F_MCP2515_reg_schreiben(BFPCTRL, 0);
	
	// set TXnRTS as inputs
	F_MCP2515_reg_schreiben(TXRTSCTRL, 0);
	
	// turn off filters => receive any message
	F_MCP2515_reg_schreiben(RXB0CTRL, (1<<RXM1) | (1<<RXM0));
	F_MCP2515_reg_schreiben(RXB1CTRL, (1<<RXM1) | (1<<RXM0));
	
	// reset device to normal mode
	F_MCP2515_reg_schreiben(CANCTRL, 0);
	
	return 1;
}


//-------------------------------------------------------------------------------------------------------
// prfen, ob neues Telegramm empfangen wurde

uint8_t F_MCP2515_check_tel(void) 
{
	return (!IS_SET(MCP2515_INT));
}


//-------------------------------------------------------------------------------------------------------
// prfen, ob ein Sendepuffer frei ist

bool F_MCP2515_freier_puffer(void)
{
	uint8_t status = F_MCP2515_status(SPI_READ_STATUS);
	
	if ((status & 0x54) == 0x54) 
  {
		// alle Puffer besetzt
		return 0;
	}
	
	return 1;
}


//-------------------------------------------------------------------------------------------------------
// empfangenes Telegramm auslesen

uint8_t F_MCP2515_tel_empfangen(tCAN *message_rec)
{
	// read status
	uint8_t status = F_MCP2515_status(SPI_RX_STATUS);
	uint8_t addr;
	
	F_SPI_Reg_MCP2515();
	
	if (bit_is_set(status,6)) 
	{
		// message in buffer 0
		addr = SPI_READ_RX;
	}
	else if (bit_is_set(status,7)) 
	{
		// message in buffer 1
		addr = SPI_READ_RX | 0x04;
	}
	else 
	{
		// Error: no message available
		return 0;
	}

	RESET(MCP2515_SS);
	F_SPI_senden(addr);
	
	// read id
	F_SPI_senden(0xFF);
	message_rec->id  = (uint16_t) F_SPI_lesen() << 3;
	F_SPI_senden(0xFF);
	message_rec->id |=            F_SPI_lesen() >> 5;
	
	F_SPI_senden(0xFF);
	F_SPI_senden(0xFF);
	
	// read DLC
	F_SPI_senden(0xFF);
	uint8_t length = F_SPI_lesen() & 0x0F;
	
	message_rec->header.length = length;
	message_rec->header.rtr = (bit_is_set(status, 3)) ? 1 : 0;
	
	// read data
	for (uint8_t i=0;i<length;i++) 
	{
		F_SPI_senden(0xFF);
		message_rec->data[i] = F_SPI_lesen();
	}

	SET(MCP2515_SS);
	
	// clear interrupt flag
	if (bit_is_set(status, 6)) 
	{
		F_MCP2515_modify(CANINTF, (1<<RX0IF), 0);
	}
	else 
	{
		F_MCP2515_modify(CANINTF, (1<<RX1IF), 0);
	}
	
	return (status & 0x07) + 1;
}


//-------------------------------------------------------------------------------------------------------
// Telegramm senden

uint8_t F_MCP2515_tel_senden(tCAN *message)
{
	uint8_t status = F_MCP2515_status(SPI_READ_STATUS);
	
	// Statusbyte:
	//
	// Bit	Function
	//  2	TXB0CNTRL.TXREQ
	//  4	TXB1CNTRL.TXREQ
	//  6	TXB2CNTRL.TXREQ

	uint8_t address;
	
	F_SPI_Reg_MCP2515();
	
	while(!F_MCP2515_freier_puffer())
	{
		// auf freien Puffer warten
	}

	if (bit_is_clear(status, 2)) // buffer 0
	{
		address = 0x00;
	}
	else if (bit_is_clear(status, 4))  // buffer 1
	{
		address = 0x02;
	} 
	else if (bit_is_clear(status, 6))  // buffer 2
	{
		address = 0x04;
	}
	else 
	{
		// all buffer used => could not send message
		return 0;
	}
	
	RESET(MCP2515_SS);
	F_SPI_senden(SPI_WRITE_TX | address);
	
	F_SPI_senden(message->id >> 3);
	F_SPI_senden(message->id << 5);
	
	F_SPI_senden(0);
	F_SPI_senden(0);
	
	uint8_t length = message->header.length & 0x0F;
	
	if (message->header.rtr) 
	{
		// a rtr-frame has a length, but contains no data
		F_SPI_senden((1<<RTR) | length);
	}
	else 
	{
		// set message length
		F_SPI_senden(length);
		
		// data
		for (uint8_t i=0;i<length;i++) 
    {
			F_SPI_senden(message->data[i]);
		}
	}

	SET(MCP2515_SS);
	
	_delay_us(1);
	
	// send message
	RESET(MCP2515_SS);
	address = (address == 0) ? 1 : address;
	F_SPI_senden(SPI_RTS | address);
	SET(MCP2515_SS);
  
	message->header.length = 8;

	// message lschen
	for(uint8_t i=0; i<8; i++)
	{
		message->data[i] = 0;
	}

	return address;
}
