//-------------------------------------------------------------------------------------------------------
//
// 1 Wire
//
//-------------------------------------------------------------------------------------------------------

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <avr/sleep.h>

#include <util/delay.h>
#include <stdio.h>
#include <stdlib.h>
#include <inttypes.h>


// Port Einstellungen

#define DS_DQ               PA8         
#define DS_PIN              PINA
#define DS_DDR              DDRA
#define T_RECOVER           2    // s




/*************************************************************************
 *
 *  Name         : BusReset
 *
 *  Beschreibung : Bus Reset und Prfen ob Sensoren angeschlossen sind
 *                 1: DQ auf 0V runterziehen und 480 us warten
 *                 2: DQ auf Eingang umschalten
 *                 3: DQ nach 10 us auf low prfen, DS=0 -> PullUp fehlt
 *                 4: 70 us auf Sensor warten, DQ auf low prfen,
 *                    DS=0 -> Sensor vorhanden, DS=1 -> kein Sensor
 *                 5: 480 us auf Sensor warten, DQ auf high prfen,
 *                    DS=1 -> Sensor O.K, DS=0 -> Bus Fehler!
 *
 *  Parameter    : keine
 *
 *  Rckgabe     : 0=O.K, 1=PullUp fehlt, 2=kein Sensor, 3=Bus Fehler
 *
 *************************************************************************/

uint8_t F_OW_BusReset(void)
{
 uint8_t Err;

 DS_DDR  |=  _BV(DS_DQ);                // DQ auf Ausgang (L)
 _delay_us(480);                        // 480 us warten
 DS_DDR &= ~_BV(DS_DQ);                 // DQ  auf Eingang (H)
 _delay_us(10);                         //  10 us warten  (auf Ext.PullUp)

 if(!(DS_PIN & _BV(DS_DQ)))             // prfe ob DQ high ist
  return(1);                            // 1 = PullUp fehlt

 _delay_us(70);                         // 70 us auf low vom Sensor warten
 Err = DS_PIN & _BV(DS_DQ) ? 2 : 0;     // 2 = kein Sensor

 _delay_us(480);                        // 480 us warten
 if(!(DS_PIN & _BV(DS_DQ)))             // prfe ob DQ wieder high ist
  return(3);                            // 3 = Bus Fehler

 return(Err);                           // 2=kein Sensor, 0=O.K
}


/*************************************************************************
 *
 *  Name         : BusWrite0Bit
 *
 *  Beschreibung : Logische 0 ausgeben
 *
 *  Parameter    : keiner
 *
 *  Rckgabe     : keiner
 *
 *************************************************************************/

void F_OW_BusWrite0Bit(void)
{
 cli();
 DS_DDR |=  _BV(DS_DQ);                 // DQ auf Ausgang (L)
 _delay_us(60);                         // 60 us Write Slot
 DS_DDR &= ~_BV(DS_DQ);                 // DQ auf Eingang (H)
 _delay_us(T_RECOVER);                  // Recovery Time
 sei();
}


/*************************************************************************
 *
 *  Name         : BusWrite1Bit
 *
 *  Beschreibung : Logische 1 ausgeben
 *
 *  Parameter    : keiner
 *
 *  Rckgabe     : keiner
 *
 *************************************************************************/

void F_OW_BusWrite1Bit(void)
{
 cli();
 DS_DDR |=  _BV(DS_DQ);                 // DQ auf Ausgang (L)
 _delay_us(T_RECOVER);                  // Recovery Time
 DS_DDR &= ~_BV(DS_DQ);                 // DQ auf Eingang (H)
 _delay_us(60);                         // 60 us Write Slot
 sei();
}


/*************************************************************************
 *
 *  Name         : BusSampleBit(void)
 *
 *  Beschreibung : DQ Lesen
 *
 *  Parameter    : keiner
 *
 *  Rckgabe     : 0/1
 *
 *************************************************************************/

uint8_t F_OW_BusSampleBit(void)
{
 uint8_t Ret;

 cli();
 DS_DDR |=  _BV(DS_DQ);                 // DQ auf Ausgang (L)
 _delay_us(T_RECOVER);                  // Recovery Time
 DS_DDR &= ~_BV(DS_DQ);                 // DQ auf Eingang (H)

 _delay_us(15);                         // 15 us warten
  Ret = DS_PIN & _BV(DS_DQ) ? 1 : 0;    // Eingang lesen
 _delay_us(45);                         // 45 us Read Slot
 sei();

 return(Ret);
}


/*************************************************************************
 *
 *  Name         : BusWriteByte
 *
 *  Beschreibung : Byte schreiben
 *
 *  Parameter    : das Byte
 *
 *  Rckgabe     : keine
 *
 *************************************************************************/

void F_OW_BusWriteByte(uint8_t Byte)
{
 uint8_t NumBytes=8;

 do{
    if(Byte & 1)                        // "1" schreiben
     F_OW_BusWrite1Bit();
    else
     F_OW_BusWrite0Bit();                    // "0" schreiben
    Byte >>= 1;
 }while(--NumBytes);
}


/*************************************************************************
 *
 *  Name         : BusReadByte
 *
 *  Beschreibung : Byte lesen
 *
 *  Parameter    : keiner
 *
 *  Rckgabe     : gelesenes Byte
 *
 *************************************************************************/

uint8_t F_OW_BusReadByte(void)
{
 uint8_t NumBytes = 8;
 uint8_t Ret = 0;
 do{
    Ret >>= 1;                          // Byte schieben
    if(F_OW_BusSampleBit())                  // Bit einlesen
     Ret |= 0x80;                       // und eintragen
 }while(--NumBytes);

 return(Ret);
}

